/*
	Copyright (C) 2008 Ivo van Doorn <IvDoorn@gmail.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <asm/byteorder.h>

#include <linux/types.h>

#include "rt2x00dump.h"

#define MINIMAL_BUFFER_LENGTH	64
#define DEFAULT_BUFFER_LENGTH	5120

static unsigned char read_dump_file = 1;

void signal_handler(int sig)
{
	signal(sig, signal_handler);

	read_dump_file = 0;
}

static inline unsigned int atoui(const char *ptr)
{
	return strtoul(ptr, NULL, 10);
}

static inline void hexdump(FILE *out, char *data, unsigned int length)
{
	unsigned int i;

	for (i = 0; i < length; i += 4) {
		if (i)
			fprintf(out, " ");
		fprintf(out, "0x%.8x", *(unsigned int *)&data[i]);
	}
}

static void dump_frame(char *frame, unsigned int read_length)
{
	struct rt2x00dump_hdr *hdr = (struct rt2x00dump_hdr *)frame;
	char *desc;
	char *data;
	unsigned int version;
	unsigned int frame_type;
	unsigned int header_length;
	unsigned int desc_length;
	unsigned int data_length;

	version = __le32_to_cpu(hdr->version);
	frame_type = __le16_to_cpu(hdr->type);

	/* Validate if we are really reading the dump file,
	 * perhaps we need to send some magic data so be
	 * absolutely sure of the validness of the data. */
	if (version != DUMP_HEADER_VERSION) {
		fprintf(stderr, "Illegal header version (%d)\n", version);
		return;
	}

	if (frame_type >= 4) {
		fprintf(stderr, "Illegal frame type detected (%d)\n", frame_type);
		return;
	}

	header_length = __le32_to_cpu(hdr->header_length);
	desc_length = __le32_to_cpu(hdr->desc_length);
	data_length = __le32_to_cpu(hdr->data_length);

	/* Validate length information based on maximum length */
	if ((header_length + desc_length) > read_length) {
		fprintf(stderr, "Read length too short for frame\n");
		return;
	} else {
		read_length -= (header_length + desc_length);
		if (read_length > data_length)
			read_length = data_length;
	}

	desc = frame + header_length;
	data = desc + desc_length;

	/* Dump rt2x00dump header */
	fprintf(stdout, "[%d][%d][%d][%d][%d][%d][%d][%d]",
		frame_type, header_length, desc_length, data_length,
		hdr->queue_index, hdr->entry_index,
		__le32_to_cpu(hdr->timestamp_sec),
		__le32_to_cpu(hdr->timestamp_usec));

	/* Dump rt2x00 descriptor */
	fprintf(stdout, "\t");
	hexdump(stdout, desc, desc_length);

	/* Dump ieee80211 frame */
	fprintf(stdout, "\t");
	hexdump(stdout, data, read_length);

	fprintf(stdout, "\n");
}

static int read_dumpfile(const char *filename, const unsigned int max_length)
{
	int rc = 0;
	int read_rc = 0;
	int file = -1;
	char *buffer = NULL;

	if (strcmp(filename, "-") == 0)
		file = STDIN_FILENO;
	else
		file = open(filename, O_RDONLY);

	if (file == -1) {
		fprintf(stderr, "Failed to open file \'%s\'\n", filename);
		rc = errno;
		goto exit;
	}

	buffer = malloc(max_length);
	if (!buffer) {
		fprintf(stderr, "Failed to allocate buffer\n");
		rc = -ENOMEM;
		goto exit;
	}

	while (read_dump_file && errno == 0) {
		memset(buffer, 0, max_length);

		read_rc = read(file, buffer, max_length);
		if (read_rc < 0)
			break;

		if (read_rc > 0)
			dump_frame(buffer, read_rc);
	}

exit:
	if (file != -1)
		close(file);

	if (buffer)
		free(buffer);

	return rc;
}

static const struct option long_options[] = {
	{ "dumpfile",	required_argument,	NULL, 'd' },
	{ "buffer",	required_argument,	NULL, 'b' },
	{ "help",	no_argument,		NULL, 'h' },
	{0, 0, 0, 0},
};

static void print_help()
{
	fprintf(stdout, "rt2x00_framedump: Debugfs TX/RX dump reader\n");
	fprintf(stdout, "Usage:\n");
	fprintf(stdout, "\trt2x00_framedump [options]\n");
	fprintf(stdout, "Options:\n");
	fprintf(stdout, "\t[-d|--dumpfile] <file>\tRead debugfs dump file\n");
	fprintf(stdout, "\t[-b|--buffer] <length>\tMax read frame length\n");
	fprintf(stdout, "\t[-h|--help]\t\tPrint help text\n");
}

int main(int argc, char *argv[])
{
	const char *filename = "-";
	unsigned int buffer_length = DEFAULT_BUFFER_LENGTH;

	while (1) {
		int opt = getopt_long(argc, argv, "b:d:h", long_options, NULL);
		if (opt == -1)
			break;

		switch (opt) {
		case 'b':
			buffer_length = atoui(optarg);
			break;
		case 'd':
			filename = optarg;
			break;
		case 'h':
			print_help();
			return;
		default:
			print_help();
			break;
		}
	}

	if (buffer_length < MINIMAL_BUFFER_LENGTH) {
		fprintf(stderr,
			"Illegal buffer length, using minimal bytes (%d)\n",
			MINIMAL_BUFFER_LENGTH);
		buffer_length = MINIMAL_BUFFER_LENGTH;
	}

	signal(SIGHUP, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGINT, signal_handler);

	if (read_dumpfile(filename, buffer_length))
		return EXIT_FAILURE;
	else
		return EXIT_SUCCESS;
}
